/*
Copyright 2016 Timo Beller, Enno Ohlebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <limits>
#include <sdsl/sdsl_concepts.hpp>
#include <sdsl/csa_wt.hpp>
#include <sdsl/suffix_array_algorithm.hpp>
#include "create_datastructures.hpp"
#include "partial_lcp.hpp"
#include "handle_graph.hpp"

using namespace std;
using namespace sdsl;
using namespace std::chrono;

struct node
{
	uint64_t lb;
	uint64_t rb;
	uint64_t len;
	bool exit_node;
	vector<uint64_t> adj_list;
	vector<uint64_t> pos_list;
	node(uint64_t _lb=0, uint64_t _rb=0, uint64_t _len=0, bool _exit_node=false) : lb(_lb), rb(_rb), len(_len), exit_node(_exit_node)
	{
	}
};

struct tmp_node
{
	uint64_t node_number;
	uint64_t lb;
	uint64_t rb;
	uint64_t len;
	tmp_node(uint64_t _node_number, uint64_t _lb=0, uint64_t _rb=0, uint64_t _len=0) : node_number(_node_number), lb(_lb), rb(_rb), len(_len)
	{
	}
};

tuple<vector<node>, vector<uint64_t>> create_cdbg_with_intervals_lin(cache_config& config, uint64_t k)
{
	// Create WT of the BWT
	typedef wt_huff<bit_vector, rank_support_v<>, select_support_scan<1>, select_support_scan<0>> wt;
	wt wt_bwt;
	construct(wt_bwt, cache_file_name(conf::KEY_BWT, config));

	// Create C-array (needed for interval_symbols)
	vector<uint64_t> carray(256, 0);
	for(uint64_t i=0, sum=0; i<256; ++i)
	{
		carray[i] = sum;
		sum += wt_bwt.rank(wt_bwt.size(), i);
	}

	// Create bit-vector of k-Intervals (marks start and end position of each k-interval with 1)
	// and add split_nodes into the graph
	vector<node> graph;
	bit_vector bv(wt_bwt.size(), 0);
	{
		int_vector<2> lcp_k = construct_partial_lcp<wt>(wt_bwt, carray, k);
		uint64_t lb=0;
		bool open=false;
		for(uint64_t i=0; i<lcp_k.size(); ++i)
		{
			if(open)
			{
				if(lcp_k[i] == lt_k)
				{
					graph.emplace_back(node(lb, i-1, k));
					bv[i-1] = true;
					open = false;
				}
			}
			else if(lcp_k[i] == eq_k)
			{
				bv[lb] = true;
				open = true;
			}
			if(lcp_k[i] == lt_k)
			{
				lb = i;
			}
		}
		if(open)
		{
			graph.emplace_back(node(lb,lcp_k.size()-1, k));
			bv[lcp_k.size()-1] = true;
		}
		// Add exit node(s)
		graph.emplace_back(node(0,0,1));
		for(uint64_t i=carray[1]; i<carray[2]; ++i)
		{
			graph.emplace_back(node(i,i,1,true));
		}
	}

	bit_vector::rank_1_type bv_rank;
	util::init_support(bv_rank, &bv);

	uint64_t quantity;
	vector<uint8_t> cs(wt_bwt.sigma);        // List of characters in the interval
	vector<uint64_t> rank_c_i(wt_bwt.sigma); // Number of occurrence of character in [0 .. i-1]
	vector<uint64_t> rank_c_j(wt_bwt.sigma); // Number of occurrence of character in [0 .. j-1]
	vector<uint64_t> start_nodes;
	for(uint64_t i=0; i<graph.size(); ++i)
	{
		tmp_node cur = tmp_node(i, graph[i].lb, graph[i].rb, graph[i].len);
		bool extend = true;
		while(extend)
		{
			extend = false;
			interval_symbols(wt_bwt, cur.lb, cur.rb+1, quantity, cs, rank_c_i, rank_c_j);
			for(uint64_t i=0; i<quantity; ++i)
			{
				uint8_t c = cs[i];
				uint64_t lb = carray[c] + rank_c_i[i];
				uint64_t rb = carray[c] + rank_c_j[i] - 1;
				uint64_t ones = bv_rank(lb+1);
				uint64_t node_number = numeric_limits<uint64_t>::max();
				//if(ones % 2 == 0 and (lb!=rb or bv[lb] == 0))
				if(ones % 2 == 0 and bv[lb] == 0)
				{
				}
				else
				{
					node_number = (ones-1)/2;
				}

				if(node_number != numeric_limits<uint64_t>::max())
				{
					graph[cur.node_number].lb = cur.lb;
					graph[cur.node_number].rb = cur.rb;
					graph[cur.node_number].len = cur.len;
				}
				else if(c<=1) // c == sentinal
				{
					graph[cur.node_number].lb = cur.lb;
					graph[cur.node_number].rb = cur.rb;
					graph[cur.node_number].len = cur.len;
					start_nodes.emplace_back(cur.node_number);
				}
				else
				{
					if(quantity == 1)
					{
						extend = true;
						cur.len++;
						cur.lb = lb;
						cur.rb = rb;
					}
					else
					{
						node_number = graph.size();
						graph.emplace_back(node(lb, rb, k));
						graph[cur.node_number].lb = cur.lb;
						graph[cur.node_number].rb = cur.rb;
						graph[cur.node_number].len = cur.len;
					}
				}
			}
		}
	}
	return make_tuple(graph, start_nodes);
}

void complete_graph(vector<node>& graph, const int_vector<>& sa)
{
	int_vector<> a(sa.size(), graph.size(), sa.width());
	uint64_t id=0;
	for(const auto& node : graph)
	{
		for(uint64_t i=node.lb; i<=node.rb; ++i)
		{
			a[sa[i]] = id;
		}
		++id;
	}
	for(uint64_t i=a.size()-1, last=a.size(); i<a.size(); --i)
	{
		if(a[i] < graph.size())
		{
			graph[a[i]].pos_list.emplace_back(i+1);
			if(last < a.size() and !graph[a[i]].exit_node) // Do not add an edge from exit node to next sequence
			{
				graph[a[i]].adj_list.emplace_back(last);
			}
			last = a[i];
		}
	}
}

int main(int argc, char *argv[])
{
	// Check parameters
	if(argc != 4)
	{
		cerr << "Usage: " << argv[0] << " inputfile outputfile kFile" << endl;
		return 1;
	}

	// Get parameters
	string inputfile = argv[1];
	string outputfile = argv[2];
	string kfilename = argv[3];

	// Create datastructures
	cache_config config(true, ".", "tmp");
	uint64_t errors = create_datastructures(config, inputfile, kfilename, true);

	// Read k-values
	ifstream kfile(kfilename);
	uint64_t k;
	while(!errors and kfile >> k)
	{
		// Create graph
		vector<node> graph;
		vector<uint64_t> start_nodes;
		tie(graph, start_nodes) = create_cdbg_with_intervals_lin(config, k);
		{	// Complete graph
			int_vector<> sa;
			load_from_file(sa, cache_file_name(conf::KEY_SA, config));
			complete_graph(graph, sa);
		}
		// Print graph
		{
			ofstream output(outputfile+".k"+to_string(k)+".dot");
			ofstream output_start_nodes(outputfile+".k"+to_string(k)+".start_nodes.txt");
			print_graph(graph, start_nodes, output, output_start_nodes);
		}
	}

	// Delete files
	if(config.delete_files)
	{
		util::delete_all_files(config.file_map);
	}

	return errors;
}
