/*
Copyright 2016 Timo Beller, Enno Ohlebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <limits>
#include <sdsl/sdsl_concepts.hpp>
#include <sdsl/csa_wt.hpp>
#include <sdsl/suffix_array_algorithm.hpp>
#include "create_datastructures.hpp"
#include "partial_lcp.hpp"
#include "handle_graph.hpp"

using namespace std;
using namespace sdsl;
using namespace std::chrono;

// WTTYPE
#ifdef WTBV
	string bvwttype = "bv";
	typedef wt_huff<bit_vector, rank_support_v<>, select_support_scan<1>, select_support_scan<0>> WT_TYPE;
#elif WTBV5
	string bvwttype = "bv5";
	typedef wt_huff<bit_vector, rank_support_v5<>, select_support_scan<1>, select_support_scan<0>> WT_TYPE;
#elif WTIL
	string bvwttype = "bvil";
	typedef wt_huff<bit_vector_il<>, rank_support_il<>, select_support_il<1>, select_support_il<0>> WT_TYPE;
#elif WTRRR15
	string bvwttype = "rrr15";
	typedef wt_huff<rrr_vector<15>, rrr_vector<15>::rank_1_type, rrr_vector<15>::select_1_type, rrr_vector<15>::select_0_type> WT_TYPE;
#elif WTRRR63
	string bvwttype = "rrr63";
	typedef wt_huff<rrr_vector<63>, rrr_vector<63>::rank_1_type, rrr_vector<63>::select_1_type, rrr_vector<63>::select_0_type> WT_TYPE;
#elif WTSD
	string bvwttype = "sd";
	typedef wt_huff<sd_vector<>, sd_vector<>::rank_1_type, sd_vector<>::select_1_type> WT_TYPE;
#else
	string bvwttype = "bv";
	typedef wt_huff<bit_vector, rank_support_v<>, select_support_scan<1>, select_support_scan<0>> WT_TYPE;
#endif

// BV1
#ifdef BV1BV
	string bv1type = "bv";
	typedef bit_vector BV1_TYPE;
#elif BV1IL
	string bv1type = "bvil";
	typedef bit_vector_il<> BV1_TYPE;
#elif BV1RRR15
	string bv1type = "rrr15";
	typedef rrr_vector<15> BV1_TYPE;
#elif BV1RRR63
	string bv1type = "rrr63";
	typedef rrr_vector<63> BV1_TYPE;
#elif BV1SD
	string bv1type = "sd";
	typedef sd_vector<> BV1_TYPE;
#else
	string bv1type = "bv";
	typedef bit_vector BV1_TYPE;
#endif

// BV2
#ifdef BV2BV
	string bv2type = "bv";
	typedef bit_vector BV2_TYPE;
#elif BV2IL
	string bv2type = "bvil";
	typedef bit_vector_il<> BV2_TYPE;
#elif BV2RRR15
	string bv2type = "rrr15";
	typedef rrr_vector<15> BV2_TYPE;
#elif BV2RRR63
	string bv2type = "rrr63";
	typedef rrr_vector<63> BV2_TYPE;
#elif BV2SD
	string bv2type = "sd";
	typedef sd_vector<> BV2_TYPE;
#else
	string bv2type = "bv";
	typedef bit_vector BV2_TYPE;
#endif

// BV3
#ifdef BV3BV
	string bv3type = "bv";
	typedef bit_vector BV3_TYPE;
#elif BV3IL
	string bv3type = "bvil";
	typedef bit_vector_il<> BV3_TYPE;
#elif BV3RRR15
	string bv3type = "rrr15";
	typedef rrr_vector<15> BV3_TYPE;
#elif BV3RRR63
	string bv3type = "rrr63";
	typedef rrr_vector<63> BV3_TYPE;
#elif BV3SD
	string bv3type = "sd";
	typedef sd_vector<> BV3_TYPE;
#else
	string bv3type = "bv";
	typedef bit_vector BV3_TYPE;
#endif



struct node
{
	uint64_t len;
	vector<uint64_t> adj_list;
	vector<uint64_t> pos_list;
	node()
	{
		len = 0;
	}
};

tuple<vector<node>, vector<uint64_t>> create_cdbg_with_lf(cache_config& config, uint64_t k)
{
	typedef WT_TYPE t_wt;
	typedef BV1_TYPE t_bv1;
	typedef BV2_TYPE t_bv2;
	typedef BV3_TYPE t_bv3;
	// Create WT of the BWT
	t_wt wt_bwt;
	construct(wt_bwt, cache_file_name(conf::KEY_BWT, config));

	// Create C-array (needed for interval_symbols)
	vector<uint64_t> carray(256, 0);
	for(uint64_t i=0, sum=0; i<256; ++i)
	{
		carray[i] = sum;
		sum += wt_bwt.rank(wt_bwt.size(), i);
	}

	// Create bit-vectors
	t_bv1 bv1;
	t_bv2 bv2;
	t_bv3 bv3;
	{
		bit_vector bv1u(wt_bwt.size(), 0);
		bit_vector bv2u(wt_bwt.size(), 0);
		bit_vector bv3u(wt_bwt.size(), 0);

		int_vector<2> lcp_k = construct_partial_lcp<t_wt>(wt_bwt, carray, k);

		int_vector_buffer<8> bwt(cache_file_name(conf::KEY_BWT, config));
		bool open=false;
		uint64_t kvalue=0;
		uint64_t lb=0;
		uint64_t last_change=0;
		vector<uint64_t> lf = carray;
		for(uint64_t i=1; i<lcp_k.size(); ++i)
		{
			++lf[bwt[i-1]];
			if(lcp_k[i] == gt_k or lcp_k[i] == eq_k)
			{
				open = true;
				if(lcp_k[i] == eq_k)
				{
					kvalue = i;
				}
			}
			else
			{
				if(open)
				{
					if(kvalue > lb)
					{
						bv1u[lb] = true;
						bv1u[i-1] = true;
					}
					if(last_change > lb)
					{
						for(uint64_t j=lb; j<=i-1; ++j)
						{
							bv2u[j] = true;
							uint8_t c = bwt[j];
							bv3u[lf[c]-1] = true;
						}
					}
					open = false;
				}
				lb = i;
			}
			if(bwt[i] != bwt[i-1] or bwt[i] <= 1)
			{
				last_change = i;
			}
		}
		if(open)
		{
			++lf[bwt[lcp_k.size()-1]];
			if(kvalue > lb)
			{
				bv1u[lb] = true;
				bv1u[lcp_k.size()-1] = true;
			}
			if(last_change > lb)
			{
				for(uint64_t j=lb; j<=lcp_k.size()-1; ++j)
				{
					bv2u[j] = true;
					uint8_t c = bwt[j];
					bv3u[lf[c]-1] = true;
				}
			}
		}
		bv3u[0] = 1;
		for(uint64_t i=1; i<carray[2]; ++i)
		{
			bv3u[i] = 0;
		}
		open = false;
		for(uint64_t i=0; i<bv1u.size(); ++i)
		{
			if(open)
			{
				bv3u[i] = 0;
				if(bv1u[i])
				{
					open = false;
				}
			}
			else if(bv1u[i])
			{
				open = true;
				bv3u[i] = 0;
			}
		}
		bv1 = t_bv1(move(bv1u));
		bv2 = t_bv2(move(bv2u));
		bv3 = t_bv3(move(bv3u));

	}

	// Init rank support and graph
	t_bv1::rank_1_type bv1_rank;
	t_bv3::rank_1_type bv3_rank;
	util::init_support(bv1_rank, &bv1);
	util::init_support(bv3_rank, &bv3);
	uint64_t number_right_max_nodes = bv1_rank(bv1.size())/2;
	vector<node> graph(number_right_max_nodes+bv3_rank(bv3.size()));
	vector<uint64_t> start_nodes;

	// Create compressed de bruijn graph
	{
		uint64_t lb = 0;
		uint64_t cur_node = number_right_max_nodes;
		graph[cur_node].len = 1;
		for(uint64_t i=wt_bwt.size()-1; i>0; --i)
		{
			// LF
			auto res = wt_bwt.inverse_select(lb);
			uint8_t c = res.second;
			uint64_t lb_new = carray[c] + res.first;

			// Right maximal
			uint64_t ones = bv1_rank(lb_new+1);
			uint64_t next_node = numeric_limits<uint64_t>::max();
			if(ones % 2 == 1 or bv1[lb_new] == 1)
			{
				next_node = (ones-1)/2;
			}

			if(c <= 1) // c == sentinal
			{
				graph[cur_node].pos_list.emplace_back(i+1);
				start_nodes.emplace_back(cur_node);
				cur_node = graph.size();
				graph.emplace_back(node());
				graph[cur_node].len = 1;
			}
			else if(next_node != numeric_limits<uint64_t>::max() or bv2[lb]) // Next node is right max or cur node is left maximal => split
			{
				if(next_node == numeric_limits<uint64_t>::max())
				{
					next_node = number_right_max_nodes + bv3_rank(lb_new);
				}
				graph[cur_node].pos_list.emplace_back(i+1);
				graph[next_node].adj_list.emplace_back(cur_node);
				graph[next_node].len = k;
				cur_node = next_node;
			}
			else
			{
				++graph[cur_node].len;
			}
			lb = lb_new;
		}
		graph[cur_node].pos_list.emplace_back(1);
		start_nodes.emplace_back(cur_node);
		reverse(begin(start_nodes), end(start_nodes));

	}
	return make_tuple(move(graph), move(start_nodes));
}

int main(int argc, char *argv[])
{
	// Check parameters
	if(argc != 4)
	{
		cerr << "Usage: " << argv[0] << " inputfile outputfile kFile" << endl;
		return 1;
	}

	// Get parameters
	string inputfile = argv[1];
	string outputfile = argv[2];
	string kfilename = argv[3];
	uint64_t binary = 0; // atoi(argv[4]);

	// Create datastructures
	cache_config config(true, ".", "tmp");
	uint64_t errors = create_datastructures(config, inputfile, kfilename, false);

	// Read k-values
	ifstream kfile(kfilename);
	uint64_t k;
	while(!errors and kfile >> k)
	{
		// Create graph
		vector<node> graph;
		vector<uint64_t> start_nodes;
		tie(graph, start_nodes) = create_cdbg_with_lf(config, k);

		// Store graph
		if(binary == 1)
		{
			string filename = outputfile+".k"+to_string(k)+".bin";
			store_graph(graph, filename);
		}

		// Print graph
		{
			ofstream output(outputfile+".k"+to_string(k)+".dot");
			ofstream output_start_nodes(outputfile+".k"+to_string(k)+".start_nodes.txt");
			print_graph(graph, start_nodes, output, output_start_nodes);
		}
	}

	// Delete files
	if(config.delete_files)
	{
		util::delete_all_files(config.file_map);
	}

	return errors;
}
