/*
Copyright 2016 Timo Beller, Enno Ohlebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <limits>
#include <sdsl/sdsl_concepts.hpp>
#include <sdsl/csa_wt.hpp>
#include <sdsl/suffix_array_algorithm.hpp>
#include <algorithm>
#include <getopt.h>
#include "create_datastructures.hpp"
#include "partial_lcp.hpp"
#include "handle_graph.hpp"

using namespace std;
using namespace sdsl;
using namespace std::chrono;

struct node
{
	uint64_t lb;
	uint64_t rb;
	uint64_t len;
	bool exit_node;
	vector<uint64_t> adj_list;
	vector<uint64_t> pos_list;
	node(uint64_t _lb=0, uint64_t _rb=0, uint64_t _len=0, bool _exit_node=false) : lb(_lb), rb(_rb), len(_len), exit_node(_exit_node)
	{
	}
};


template<class t_wt=wt_huff<bit_vector, rank_support_v<>, select_support_mcl<1>, select_support_mcl<0>>,
         class t_bv1=bit_vector,
         class t_bv3=bit_vector,
	 class t_wt_doc=wt_huff<bit_vector, rank_support_v<>, select_support_scan<1>, select_support_scan<0>, int_tree<>>
	 >
struct compressed_debruijn_graph
{
	typedef uint64_t size_type;

	private:
	struct node_c
	{
		uint64_t lb;       // lb of prefix kmer
		uint64_t len;
		uint64_t size;
		uint64_t first_lb; // lb of suffix kmer
		node_c(uint64_t _lb=0, uint64_t _len=0, uint64_t _size=0, uint64_t _first_lb=0) : lb(_lb), len(_len), size(_size), first_lb(_first_lb)
		{
		}
	};

	uint64_t                    m_k;
	t_wt                        m_wt_bwt;
	vector<uint64_t>            m_carray;
	vector<node_c>              m_nodes;
	uint64_t                    m_right_max;
	vector<uint64_t>	    m_stop_nodes;
	t_bv1                       m_bv1;
	t_bv3                       m_bv3;
	typename t_bv1::rank_1_type m_bv1_rank;
	typename t_bv3::rank_1_type m_bv3_rank;
	t_wt_doc                    m_wt_doc;


	void create_carray()
	{
		m_carray = vector<uint64_t>(256, 0);
		for(uint64_t i=0, sum=0; i<256; ++i)
		{
			m_carray[i] = sum;
			sum += m_wt_bwt.rank(m_wt_bwt.size(), i);
		}
	}

	void detect_nodes(cache_config& config)
	{
		// Create int_vector<2> that indicates if the lcp value is smaller, eqal or greater than k
		int_vector<2> lcp_k = construct_partial_lcp<t_wt>(m_wt_bwt, m_carray, m_k);

		// It should be possible to stream lcp_k from file ... but memory peak is during construct_partial_lcp
		stack<uint64_t> indexes;
		bit_vector bv1(m_wt_bwt.size(), 0);
		bit_vector bv3(m_wt_bwt.size(), 0);

		// Add right_maximal nodes
		bool open=false;
		uint64_t kvalue=0;
		uint64_t lb=0;
		uint64_t last_change=0;
		vector<uint64_t> lf = m_carray;
		int_vector_buffer<8> bwt(cache_file_name(conf::KEY_BWT, config));
		for(uint64_t i=1; i<lcp_k.size(); ++i)
		{
			++lf[bwt[i-1]];
			if(lcp_k[i] == gt_k or lcp_k[i] == eq_k)
			{
				open = true;
				if(lcp_k[i] == eq_k)
				{
					kvalue = i;
				}
			}
			else
			{
				if(open)
				{
					if(kvalue > lb)
					{
						bv1[lb] = true;
						bv1[i-1] = true;
						m_nodes.emplace_back(node_c(lb, m_k, i-lb, lb));
					}
					if(last_change > lb)
					{
						for(uint64_t j=lb; j<=i-1; ++j)
						{
							uint8_t c = bwt[j];
							bv3[lf[c]-1] = true;
						}
					}
					open = false;
				}
				lb = i;
			}
			if(bwt[i] != bwt[i-1] or bwt[i] <= 1)
			{
				last_change = i;
			}
		}
		if(open)
		{
			++lf[bwt[lcp_k.size()-1]];
			if(kvalue > lb)
			{
				bv1[lb] = true;
				bv1[lcp_k.size()-1] = true;
				m_nodes.emplace_back(node_c(lb, m_k, lcp_k.size()-lb, lb));
			}
			if(last_change > lb)
			{
				for(uint64_t j=lb; j<=lcp_k.size()-1; ++j)
				{
					uint8_t c = bwt[j];
					bv3[lf[c]-1] = true;
				}
			}
		}

		// Add Endnodes
		for(uint64_t i=0; i<m_carray[2]; ++i)
		{
			m_stop_nodes.emplace_back(m_nodes.size());
			m_nodes.emplace_back(node_c(i, 1, 1, i));
			bv3[i] = 0;
		}

		open = false;
		for(uint64_t i=0; i<bv1.size(); ++i)
		{
			if(open)
			{
				bv3[i] = 0;
				if(bv1[i])
				{
					open = false;
				}
			}
			else if(bv1[i])
			{
				bv3[i] = 0;
				open = true;
			}
		}

		// Create rank support
		m_bv1 = t_bv1(move(bv1));
		m_bv3 = t_bv3(move(bv3));

		util::init_support(m_bv1_rank, &m_bv1);
		util::init_support(m_bv3_rank, &m_bv3);
	}

	void complete_nodes()
	{
		uint64_t quantity;
		vector<uint8_t> cs(m_wt_bwt.sigma);        // List of characters in the interval
		vector<uint64_t> rank_c_i(m_wt_bwt.sigma); // Number of occurrence of character in [0 .. i-1]
		vector<uint64_t> rank_c_j(m_wt_bwt.sigma); // Number of occurrence of character in [0 .. j-1]
		stack<uint64_t> order;
		for(uint64_t i=0, undef=numeric_limits<uint64_t>::max(); i<m_nodes.size(); ++i)
		{
			// Determine which node should be completed next
			uint64_t nodeid = i;
			if(i >= m_right_max)
			{
				nodeid = order.top();
				order.pop();
			}

			uint64_t cur_lb = m_nodes[nodeid].lb;
			uint64_t cur_rb = m_nodes[nodeid].lb+m_nodes[nodeid].size-1;
			uint64_t cur_len = m_nodes[nodeid].len;
			bool extend = true;
			while(extend)
			{
				extend = false;
				interval_symbols(m_wt_bwt, cur_lb, cur_rb+1, quantity, cs, rank_c_i, rank_c_j);
				for(uint64_t j=0; j<quantity; ++j)
				{
					uint8_t c = cs[j];
					uint64_t lb = m_carray[c] + rank_c_i[j];
					uint64_t rb = m_carray[c] + rank_c_j[j] - 1;
					uint64_t ones = m_bv1_rank(lb+1);
					uint64_t node_number = undef;

					if(ones % 2 == 0 and m_bv1[lb] == 0)
					{
					}
					else
					{
						node_number = (ones-1)/2;
					}

					if(node_number != undef)
					{
						m_nodes[nodeid].lb = cur_lb;
						m_nodes[nodeid].len = cur_len;
					}
					else if(c<=1) // c == sentinal
					{
						m_nodes[nodeid].lb = cur_lb;
						m_nodes[nodeid].len = cur_len;
					}
					else
					{
						if(quantity == 1)
						{
							extend = true;
							cur_len++;
							cur_lb = lb;
							cur_rb = rb;
						}
						else
						{
							uint64_t next_node_id = m_right_max + m_bv3_rank(lb);

							order.push(next_node_id);
							m_nodes[next_node_id] = node_c(lb, m_k, rb-lb+1, lb);

							m_nodes[nodeid].lb = cur_lb;
							m_nodes[nodeid].len = cur_len;
						}
					}
				}
			}
		}
	}

	public:

	compressed_debruijn_graph() {}

	compressed_debruijn_graph(cache_config& config, uint64_t k, bool with_document_array) : m_k(k)
	{
		// Create WT of the BWT
		construct(m_wt_bwt, cache_file_name(conf::KEY_BWT, config));

		// Create C-array (needed for interval_symbols)
		create_carray();

		// Detect and create nodes incl. bit vectors for calculation node numbers
		detect_nodes(config);

		// Add space for nodes not ending with an right maximal kmer
		m_right_max = m_nodes.size();
		uint64_t lmax = m_bv3_rank(m_bv3.size());
		m_nodes.resize(m_right_max + lmax);

		// Complete nodes
		complete_nodes();

		// Load Document Array
		if(with_document_array)
		{
			construct(m_wt_doc, cache_file_name("DA", config));
		}
	}


	tuple<vector<node>,vector<uint64_t>> get_explicit_representation() const
	{
		vector<node> graph(m_nodes.size());
		uint64_t d = m_carray[2];
		vector<uint64_t> start_nodes(d);

		uint64_t pos = m_wt_bwt.size()+1; // 1 indexed
		for(uint64_t s=0, i=0; s<d; ++s)
		{
			uint64_t prev_node_number = m_right_max-m_carray[2]+i;

			uint64_t idx = m_nodes[prev_node_number].lb;
			pos -= m_nodes[prev_node_number].len;

			graph[prev_node_number].len = m_nodes[prev_node_number].len;
			graph[prev_node_number].pos_list.emplace_back(pos);
			auto res = m_wt_bwt.inverse_select(idx);
			i = m_carray[res.second] + res.first;
			while(res.second > 1) // c != sentinal
			{
				uint64_t ones = m_bv1_rank(i+1);
				uint64_t node_number = 0;
				if(ones % 2 == 0 and m_bv1[i] == 0)
				{
					node_number = m_right_max + m_bv3_rank(i);
				}
				else
				{
					node_number = (ones-1)/2;
				}

				idx = m_nodes[node_number].lb + (i-m_nodes[node_number].first_lb);
				pos -= (m_nodes[node_number].len-m_k+1);

				// Store all information in new graph
				graph[node_number].adj_list.emplace_back(prev_node_number);
				graph[node_number].pos_list.emplace_back(pos);
				graph[node_number].len = m_nodes[node_number].len;

				// Go Node back
				prev_node_number = node_number;
				res = m_wt_bwt.inverse_select(idx);
				i = m_carray[res.second] + res.first;
			}
			start_nodes[d-1-s] = prev_node_number;
		}
		return make_tuple(move(graph), move(start_nodes));
	}

	void print_graph_statistics(ostream& out) const
	{
		// Print graph overview
		{
			uint64_t labels = 0;
			uint64_t edges = 0;
			uint64_t repeat_nodes = 0;
			uint64_t unique_nodes = 0;

			for(const auto& node : m_nodes)
			{
				labels += node.size;
				edges += node.size;
				if(node.size == 1)
				{
					++unique_nodes;
				}
				else
				{
					++repeat_nodes;
				}
			}
			edges -= m_stop_nodes.size(); // Stop nodes have no outgoing edge
			out << "       nodes=" << setw(10) << m_nodes.size() << endl;
			out << "unique nodes=" << setw(10) << unique_nodes << endl;
			out << "repeat nodes=" << setw(10) << repeat_nodes << endl;
			out << "      labels=" << setw(10) << labels << endl;
			out << "       edges=" << setw(10) << edges << endl;
			out << endl;
		}

		// Print lengths - statistics
		{
			uint64_t sum_length = 0;
			uint64_t sum_length_unique_nodes = 0;
			uint64_t sum_length_repeat_nodes = 0;
			uint64_t unique_nodes = 0;
			uint64_t repeat_nodes = 0;
			uint64_t max_length = 0;
			for(const auto& node: m_nodes)
			{
				sum_length += node.len;
				if(node.size == 1)
				{
					++unique_nodes;
					sum_length_unique_nodes += node.len;
				}
				else
				{
					++repeat_nodes;
					sum_length_repeat_nodes += node.len;
				}
				if(max_length < node.len)
				{
					max_length = node.len;
				}
			}
			out << "max_length=" << max_length << endl;
			out << "avg_length=" << sum_length/(double)m_nodes.size() << endl;
			out << "avg_length(unique_nodes)=" << sum_length_unique_nodes/(double)unique_nodes << endl;
			out << "avg_length(repeat_nodes)=" << sum_length_repeat_nodes/(double)repeat_nodes << endl;
			if(false)
			{
				vector<uint64_t> lengths(101, 0);
				uint64_t diff = (max_length+1)/(lengths.size()-1);
				if(diff == 0)
				{
					diff = 1;
				}
				for(const auto& node: m_nodes)
				{
					++lengths[node.len/diff];
				}
				for(uint64_t i=0; i<lengths.size(); ++i)
				{
					out << setw(10) << (i*diff) << " - " << setw(10) << ((i+1)*diff) << ": " << lengths[i] << " nodes" << endl;
				}
			}
			out << endl;
			out << "==========================================================================" << endl;
			out << endl;
		}

		// Print node-to-sequence relation details
		{
			uint64_t seq_number = m_wt_doc.sigma;

			// k = node_pair[i][j] means:
			// There are k nodes that contain (at least) sequence i and j
			vector<vector<uint64_t>> node_pair(seq_number, vector<uint64_t>(seq_number, 0));

			/// k = count_node_sequences[j] means:
			// There are k nodes that contain exactly j sequences
			vector<uint64_t> count_node_sequnces(seq_number+1, 0);

			uint64_t quantity;
			vector<uint64_t> cs(m_wt_doc.sigma);       // List of sequences in the interval
			vector<uint64_t> rank_c_i(m_wt_doc.sigma); // Number of occurrence of character in [0 .. i-1]
			vector<uint64_t> rank_c_j(m_wt_doc.sigma); // Number of occurrence of character in [0 .. j-1]
			for(const auto& node : m_nodes)
			{
				interval_symbols(m_wt_doc, node.lb, node.lb+node.size, quantity, cs, rank_c_i, rank_c_j);
				++count_node_sequnces[quantity];
				for(uint64_t i=0; i<quantity; ++i)
				{
					uint64_t seq_a = cs[i];
					for(uint64_t j=i; j<quantity; ++j)
					{
						uint64_t seq_b = cs[j];
						++node_pair[seq_a][seq_b];
					}
				}
			}


			// "Sort"
			for(uint64_t i=0; i<node_pair.size(); ++i)
			{
				for(uint64_t j=i+1; j<node_pair[i].size(); ++j)
				{
					node_pair[i][j] += node_pair[j][i];
					node_pair[j][i] = 0;
				}
			}

			// Print Results
			for(uint64_t i=0; i<count_node_sequnces.size(); ++i)
			{
				out << setw(10) << count_node_sequnces[i] << " nodes covers exactly " << setw(3) << i << " sequences." << endl;
			}
			out << endl;
			for(uint64_t i=0; i<node_pair.size(); ++i)
			{
				for(uint64_t j=0; j<node_pair[i].size(); ++j)
				{
					out << setw(11) << node_pair[i][j];
				}
				out << endl;
			}
			out << endl;
		}
	}

	uint64_t get_k() const
	{
		return m_k;
	}

	vector<uint64_t> get_stop_nodes() const
	{
		return m_stop_nodes;
	}

	// Find all nodes that contains the pattern s
	tuple<vector<uint64_t>,uint64_t> find_nodes(const string& s) const
	{
		vector<uint64_t> result;
		assert(s.size() >= m_k);
		if(s.size() < m_k)
		{
			cerr << "Pattern length must at least " << m_k << endl;
			return make_tuple(result, -1);
		}

		// Find sa-interval
		uint64_t i = 0;
		uint64_t j = m_wt_bwt.size()-1;
		uint64_t pos = s.size();
		while(i<j and pos > s.size()-m_k)
		{
			pos--;
			uint8_t c = s[pos];
			i = m_carray[c] + m_wt_bwt.rank(i  , c);
			j = m_carray[c] + m_wt_bwt.rank(j+1, c)-1;
		}
		while(i<=j and pos > s.size()-m_k)
		{
			pos--;
			uint8_t c = s[pos];
			auto res = m_wt_bwt.inverse_select(i);
			if(res.second == c)
			{
				i = m_carray[c] + res.first;
				j = i;
			}
			else
			{
				i = j+1; // Not Found!
			}
	}
		if(i > j)
		{
			return make_tuple(result, -1);
		}

		uint64_t lb = i;
		uint64_t rb = j;

		// Find nodeid of end node (node that contains the suffix of length k)
		uint64_t nodeid = m_wt_bwt.size()+1;
		uint64_t l = 0;
		while(nodeid > m_wt_bwt.size())
		{
			uint64_t ones = m_bv1_rank(i+1);
			if(ones % 2 == 1 or m_bv1[i] == 1)
			{
				nodeid = (ones-1)/2;
			}
			else
			{
				uint64_t ones_i = m_bv3_rank(i);
				uint64_t ones_j = m_bv3_rank(j+1);
				if(ones_i != ones_j)
				{
					nodeid = m_right_max + ones_i;
				}
				else
				{
					uint8_t c = 0;
					if(l < m_k)
					{
						c = s[s.size()-m_k+l];
					}
					else
					{
						while(i >= m_carray[c])
						{
							++c;
						}
						--c;
					}

					if(i == j)
					{
						i = m_wt_bwt.select(i-m_carray[c]+1, c); // ilf
						j = i;
					}
					else
					{
						i = m_wt_bwt.select(i-m_carray[c]+1, c); // ilf
						j = m_wt_bwt.select(j-m_carray[c]+1, c); // ilf
					}
					++l;
					if(i < m_carray[2]) // Found sentinal
					{
						nodeid = m_right_max-m_carray[2]+i;
						l -= (m_k-1);
					}
				}
			}
		}
		l = m_nodes[nodeid].len - l - m_k; // Start position of suffix in current node

		// Add nodeid to path
		result.emplace_back(nodeid);

		// Find preceeding nodes
		i = lb;
		j = rb;
		while(i<=j and pos > 0)
		{
			// Add character to the front of the suffix
			pos--;
			uint8_t c = s[pos];
			i = m_carray[c] + m_wt_bwt.rank(i  , c);
			j = m_carray[c] + m_wt_bwt.rank(j+1, c)-1; // if i == j this can be done better with inverse_select!
			// auto res = m_wt_bwt.inverse_select(i);

			// Check if I'm in a new node
			if(l == 0)
			{
				uint64_t ones = m_bv1_rank(i+1);
				if(ones % 2 == 1 or m_bv1[i] == 1)
				{
					nodeid = (ones-1)/2;
				}
				else
				{
					nodeid = m_right_max + m_bv3_rank(i);
				}


				l = m_nodes[nodeid].len - m_k;
			}
			else
			{
				--l;
			}
			// Add nodeid to path
			result.emplace_back(nodeid);
		}

		// If pattern was not found
		if(i>j)
		{
			result.resize(0);
			l = 0;
		}

		return make_tuple(move(result), l);
	}

	// Find all sequences that occur in a node
	// Precondition: nodeid is valid
	vector<uint64_t> sequences_in_node(const uint64_t nodeid) const
	{
		assert(nodeid < m_nodes.size());
		auto node = m_nodes[nodeid];
		vector<uint64_t> result;
		uint64_t quantity;
		vector<uint64_t> cs(m_wt_doc.sigma);       // List of sequences in the interval
		vector<uint64_t> rank_c_i(m_wt_doc.sigma); // Number of occurrence of character in [0 .. i-1]
		vector<uint64_t> rank_c_j(m_wt_doc.sigma); // Number of occurrence of character in [0 .. j-1]
		interval_symbols(m_wt_doc, node.lb, node.lb+node.size, quantity, cs, rank_c_i, rank_c_j);
		for(uint64_t i=0; i<quantity; ++i)
		{
			result.emplace_back(cs[i]);
		}

		return result;
	}


	template<class t_pod>
	uint64_t serialize_vector_pod(const vector<t_pod>& vpod, std::ostream& out, structure_tree_node* v, std::string name="") const
	 {
		 structure_tree_node* child = structure_tree::add_child(v, name, "vector<POD>");
		 uint64_t written_bytes = 0;
		 uint64_t size=vpod.size();
		 written_bytes += write_member(size, out);
		 out.write((char*)vpod.data(), vpod.size()*sizeof(vpod[0]));
		 written_bytes += vpod.size()*sizeof(vpod[0]);
		 structure_tree::add_size(child, written_bytes);
		 return written_bytes;
	 }

	template<class t_pod>
	void load_vpod(vector<t_pod>& vpod, std::istream& in)
	{
		uint64_t size;
		read_member(size, in);
		vpod.resize(size);
		in.read((char*)vpod.data(), size*sizeof(vpod[0]));
	}



         //! Serialize into stream
         size_type serialize(std::ostream& out, structure_tree_node* v=nullptr, std::string name="") const
	 {
		 structure_tree_node* child = structure_tree::add_child(v, name, util::class_name(*this));
		 size_type written_bytes = 0;
		 written_bytes += write_member(m_k, out, child, "k");
		 written_bytes += m_wt_bwt.serialize(out, child, "wt_bwt");
		 written_bytes += serialize_vector_pod(m_carray, out, child, "c_array");
		 written_bytes += serialize_vector_pod(m_nodes, out, child, "nodes");
		 written_bytes += write_member(m_right_max, out, child, "right_max");
		 written_bytes += serialize_vector_pod(m_stop_nodes, out, child, "stop_nodes");
		 written_bytes += m_bv1.serialize(out, child, "bv1");
		 written_bytes += m_bv3.serialize(out, child, "bv3");
		 written_bytes += m_bv1_rank.serialize(out, child, "bv1_rank");
		 written_bytes += m_bv3_rank.serialize(out, child, "bv3_rank");
		 written_bytes += m_wt_doc.serialize(out, child, "wt_doc");

		 structure_tree::add_size(child, written_bytes);
		 return written_bytes;
	 }

         //! Load sampling from disk
         void load(std::istream& in)
	 {
		 read_member(m_k, in);
		 m_wt_bwt.load(in);
		 load_vpod(m_carray, in);
		 load_vpod(m_nodes, in);
		 read_member(m_right_max, in);
		 load_vpod(m_stop_nodes, in);
		 m_bv1.load(in);
		 m_bv3.load(in);
		 m_bv1_rank.load(in, &m_bv1);
		 m_bv3_rank.load(in, &m_bv3);
		 m_wt_doc.load(in);
	 }
};


// WTTYPE
#ifdef WTBV
	string bvwttype = "bv";
	typedef wt_huff<bit_vector, rank_support_v<>, select_support_mcl<1>, select_support_mcl<0>> WT_TYPE;
#elif WTBV5
	string bvwttype = "bv5";
	typedef wt_huff<bit_vector, rank_support_v5<>, select_support_mcl<1>, select_support_mcl<0>> WT_TYPE;
#elif WTIL
	string bvwttype = "bvil";
	typedef wt_huff<bit_vector_il<>, rank_support_il<>, select_support_il<1>, select_support_il<0>> WT_TYPE;
#elif WTRRR15
	string bvwttype = "rrr15";
	typedef wt_huff<rrr_vector<15>, rrr_vector<15>::rank_1_type, rrr_vector<15>::select_1_type, rrr_vector<15>::select_0_type> WT_TYPE;
#elif WTRRR63
	string bvwttype = "rrr63";
	typedef wt_huff<rrr_vector<63>, rrr_vector<63>::rank_1_type, rrr_vector<63>::select_1_type, rrr_vector<63>::select_0_type> WT_TYPE;
#elif WTSD
	string bvwttype = "sd";
	typedef wt_huff<sd_vector<>, sd_vector<>::rank_1_type, sd_vector<>::select_1_type> WT_TYPE;
#else
	string bvwttype = "bv";
	typedef wt_huff<bit_vector, rank_support_v<>, select_support_mcl<1>, select_support_mcl<0>> WT_TYPE;
#endif

// BV1
#ifdef BV1BV
	string bv1type = "bv";
	typedef bit_vector BV1_TYPE;
#elif BV1IL
	string bv1type = "bvil";
	typedef bit_vector_il<> BV1_TYPE;
#elif BV1RRR15
	string bv1type = "rrr15";
	typedef rrr_vector<15> BV1_TYPE;
#elif BV1RRR63
	string bv1type = "rrr63";
	typedef rrr_vector<63> BV1_TYPE;
#elif BV1SD
	string bv1type = "sd";
	typedef sd_vector<> BV1_TYPE;
#else
	string bv1type = "bv";
	typedef bit_vector BV1_TYPE;
#endif

// BV3
#ifdef BV3BV
	string bv3type = "bv";
	typedef bit_vector BV3_TYPE;
#elif BV3IL
	string bv3type = "bvil";
	typedef bit_vector_il<> BV3_TYPE;
#elif BV3RRR15
	string bv3type = "rrr15";
	typedef rrr_vector<15> BV3_TYPE;
#elif BV3RRR63
	string bv3type = "rrr63";
	typedef rrr_vector<63> BV3_TYPE;
#elif BV3SD
	string bv3type = "sd";
	typedef sd_vector<> BV3_TYPE;
#else
	string bv3type = "bv";
	typedef bit_vector BV3_TYPE;
#endif

typedef compressed_debruijn_graph<WT_TYPE, BV1_TYPE, BV3_TYPE> CDBG;

void find_pattern(const string& filename_graph, const string& filename_pattern, bool verbose)
{
	// Load Graph an input
	CDBG g;
	load_from_file(g, filename_graph);
	ifstream patternfile(filename_pattern);

	string p;
	vector<string> pattern;
	auto t1 = high_resolution_clock::now();
	auto t2 = high_resolution_clock::now();
	auto time_search_pattern = t2-t2;
	auto time_documents_of_first_node = t2-t2;
	auto time_documents_of_all_nodes = t2-t2;

	uint64_t number_patterns = 0;
	uint64_t number_found = 0;

	if(verbose)
	{
		while(patternfile >> p)
		{
			++number_patterns;
			vector<uint64_t> node_sequences;
			uint64_t tmp;
			t1 = high_resolution_clock::now();
			tie(node_sequences, tmp) = g.find_nodes(p);
			t2 = high_resolution_clock::now();
			time_search_pattern += t2-t1;

			if(node_sequences.size())
			{
				cout << "Pattern '" << p << "' occurs in the following nodes: ";
				cout << node_sequences[0];
				for(uint64_t i=1; i<node_sequences.size(); ++i)
				{
					cout << ", " << node_sequences[i];
				}
				cout << endl;
				++number_found;
				t1 = high_resolution_clock::now();
				vector<uint64_t> seq = g.sequences_in_node(node_sequences.front());
				t2 = high_resolution_clock::now();
				time_documents_of_first_node += t2-t1;

				t1 = high_resolution_clock::now();
				for(const auto& nodeid : node_sequences)
				{
					vector<uint64_t> seq = g.sequences_in_node(nodeid);
					cout << "Node " << nodeid << " corresponds to a substring that occurs in the following sequences: ";
					cout << seq[0];
					for(uint64_t i=1; i<seq.size(); ++i)
					{
						cout << ", " << seq[i];
					}
					cout << endl;
				}
				t2 = high_resolution_clock::now();
				time_documents_of_all_nodes += t2-t1;
			}
			else
			{
				cout << "Pattern '" << p << "' does not occur." << endl;
			}
			cout << endl;
		}
		cout << "Found " << number_found << " of " << number_patterns << " Pattern" << endl;
	}
	else
	{
		while(patternfile >> p)
		{
			++number_patterns;
			vector<uint64_t> node_sequences;
			uint64_t tmp;
			t1 = high_resolution_clock::now();
			tie(node_sequences, tmp) = g.find_nodes(p);
			t2 = high_resolution_clock::now();
			time_search_pattern += t2-t1;

			if(node_sequences.size())
			{
				++number_found;
				t1 = high_resolution_clock::now();
				vector<uint64_t> seq = g.sequences_in_node(node_sequences.front());
				t2 = high_resolution_clock::now();
				time_documents_of_first_node += t2-t1;

				t1 = high_resolution_clock::now();
				for(const auto& nodeid : node_sequences)
				{
					vector<uint64_t> seq = g.sequences_in_node(nodeid);
				}
				t2 = high_resolution_clock::now();
				time_documents_of_all_nodes += t2-t1;
			}
		}
	}
	cout << setw(10) << duration_cast<milliseconds>(time_search_pattern).count() << "ms to search pattern." << endl;
	cout << setw(10) << duration_cast<milliseconds>(time_documents_of_first_node).count() << "ms to list documents of the first node." << endl;
	cout << setw(10) << duration_cast<milliseconds>(time_documents_of_all_nodes).count() << "ms to list documents of all nodes." << endl;
}

bool convert_impl2expl(string& filename_graph, string& filename_output)
{
	CDBG g;
	load_from_file(g, filename_graph);

	vector<node> graph;
	vector<uint64_t> start_nodes;
	tie(graph, start_nodes) = g.get_explicit_representation();

	ofstream output(filename_output+".dot");
	ofstream output_start_nodes(filename_output+".start_nodes.txt");
	if(!output.is_open() or !output_start_nodes.is_open())
	{
		cerr << "Could not open '" << filename_output << ".dot' or '";
		cerr << filename_output << ".start_nodes.txt' for writing." << endl;
		return false;
	}
	else
	{
		print_graph(graph, start_nodes, output, output_start_nodes);
	}
	return true;
}

void construct(const string& inputfile, const string& outputfile, const string& kfilename, bool with_document_array, bool write_dot_file)
{
	uint64_t min_length = 0;

	// Create datastructures
	cache_config config(true, ".", "tmp");
	{
		// Get input
		vector<uint64_t> sequences = create_text(config, inputfile, true);
		auto min = min_element(begin(sequences), end(sequences));
		if(min!=end(sequences))
		{
			min_length = *min;
		}

		// Get sa
		bool fast = false; // space-efficient!
		create_sa(config, fast);

		// Get bwt
		create_bwt(config);

		// Get document array
		if(with_document_array)
		{
			create_da(config, sequences);
		}
	}

	// Read k-values
	ifstream kfile(kfilename);
	uint64_t k;
	while(kfile >> k)
	{
		if(min_length < k)
		{
			cerr << "k=" << k << " must smaller than sequence length";
			cerr << ", but in input file '" << inputfile << "' there is a sequence with length ";
			cerr << min_length << " - this k-values will be skipped." << endl;
		}
		else
		{
			// Create graph
			vector<node> graph;
			vector<uint64_t> start_nodes;
			{
				CDBG g(config, k, with_document_array);

				// Store graph
				if(write_dot_file)
				{
					// Convert graph
					tie(graph,start_nodes) = g.get_explicit_representation();

					ofstream output(outputfile+".k"+to_string(k)+".dot");
					ofstream output_start_nodes(outputfile+".k"+to_string(k)+".start_nodes.txt");
					print_graph(graph, start_nodes, output, output_start_nodes);
				}
				else
				{
					ofstream out(outputfile+".k"+to_string(k)+".bin");
					g.serialize(out);
				}
			}
		}
	}

	// Delete files
	if(config.delete_files)
	{
		util::delete_all_files(config.file_map);
	}
}

void usage(string argv_null, const string& command)
{
	if(command == "construct")
	{
		cerr << "usage: " << argv_null << " " << command << " options" << endl;
		cerr << endl;
		cerr << setw(10) << "-i, --inputfile=INFILE" << " the input file" << endl;
		cerr << setw(10) << "-o, --outputfile=OUTFILE" << " the output file" << endl;
		cerr << setw(10) << "-k, --kfile=KFILE" << " text file containing k values" << endl;
	}
	else if(command == "print_graph_details")
	{
		cerr << "usage: " << argv_null << " " << command << " options" << endl;
		cerr << endl;
		cerr << setw(10) << "-g, --graphfile=GRAPHFILE" << " graph file, created via construct command" << endl;
	}
	else if(command == "find_pattern")
	{
		cerr << "usage: " << argv_null << " " << command << " options" << endl;
		cerr << endl;
		cerr << setw(10) << "-g, --graphfile=GRAPHFILE" << " graph file, created via construct command" << endl;
		cerr << setw(10) << "-p, --patternfile=PATTERNFILE" << " pattern file, containing pattern" << endl;
	}
	else if(command == "impl2expl")
	{
		cerr << "usage: " << argv_null << " " << command << " options" << endl;
		cerr << "Program will create OUTFILE.dot and OUTFILE.start_nodes.txt" << endl;
		cerr << endl;
		cerr << setw(10) << "-g, --graphfile=GRAPHFILE" << " graph file, created via construct command" << endl;
		cerr << setw(10) << "-o, --outputfile=OUTFILE" << " the output file" << endl;
	}
	else
	{
		cerr << "usage: " << argv_null << " <command> <args>" << endl;
		cerr << endl;
		cerr << "Where command must be one of the following list:" << endl;
		cerr << setw(25) << left << "   construct" << " - Construct the compressed de bruijn graph" << endl;
		cerr << setw(25) << left << "   print_graph_details" << " - Print graph details" << endl;
		cerr << setw(25) << left << "   find_pattern" << " - Finding pattern in the pan-genome" << endl;
		cerr << setw(25) << left << "   impl2expl" << " - Convert to explicit representation" << endl;
	}
	cerr << endl;
}

void check_argument_given(string argv_null, string& command, string& value, string name)
{
	if(value == "")
	{
		usage(argv_null, command);
		cerr << "ERROR: No " << name << " given." << endl;
		exit(EXIT_FAILURE);
	}
}

int main (int argc, char **argv)
{
	if(argc <= 1)
	{
		usage(argv[0], "");
		exit(1);
	}
	string command = string(argv[1]);

	static struct option long_options[] =
	{
		{"inputfile",   required_argument, nullptr, 'i'},
		{"outputfile",  required_argument, nullptr, 'o'},
		{"kfile",       required_argument, nullptr, 'k'},
		{"graphfile",   required_argument, nullptr, 'g'},
		{"patternfile", required_argument, nullptr, 'p'},
		{nullptr, 0, nullptr, 0}
	};

	string inputfile;
	string outputfile;
	string kfile;
	string graphfile;
	string patternfile;

	int64_t c;
	while ((c = getopt_long (argc, argv, "i:o:k:g:p:", long_options, nullptr) ) != -1)
	{
		switch (c)
		{
			case 'i':
				inputfile = string(optarg);
				break;

			case 'o':
				outputfile = string(optarg);
				break;

			case 'k':
				kfile = string(optarg);
				break;

			case 'g':
				graphfile = string(optarg);
				break;

			case 'p':
				patternfile = string(optarg);
				break;

			default:
				usage(argv[0], command);
				exit(1);
		}
	}

	if(command == "construct")
	{
		check_argument_given(argv[0], command, inputfile, "inputfile");
		check_argument_given(argv[0], command, outputfile, "outputfile");
		check_argument_given(argv[0], command, kfile, "kfile");
		construct(inputfile, outputfile, kfile, true, false);
	}
	else if(command == "print_graph_details")
	{
		check_argument_given(argv[0], command, graphfile, "graphfile");

		cerr << endl << graphfile<< ":" << endl;
		CDBG g;
		load_from_file(g, graphfile);
		g.print_graph_statistics(cerr);
	}
	else if(command == "find_pattern")
	{
		check_argument_given(argv[0], command, graphfile, "graphfile");
		check_argument_given(argv[0], command, patternfile, "patternfile");
		find_pattern(graphfile, patternfile, true);
	}
	else if(command == "impl2expl")
	{
		check_argument_given(argv[0], command, graphfile, "graphfile");
		check_argument_given(argv[0], command, outputfile, "outputfile");
		if(!convert_impl2expl(graphfile, outputfile))
		{
			exit(1);
		}
	}
	else
	{
		usage(argv[0], command);
		exit(1);
	}

	exit (0);
}
