/*
Copyright 2016 Timo Beller, Enno Ohlebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef INCLUDED_CREATE_DATASTRUCTURES
#define INCLUDED_CREATE_DATASTRUCTURES

#include <iostream>
#include <fstream>
#include <limits>
#include <algorithm>
#include <sdsl/sdsl_concepts.hpp>
#include <sdsl/suffix_array_algorithm.hpp>

using namespace std;
using namespace sdsl;
using namespace std::chrono;


vector<uint64_t> create_text(cache_config& config, const string& inputfile, bool calc_sequences=true)
{
	vector<uint64_t> sequences;

	// (1) Check, if the text is cached
	if ( !cache_file_exists(conf::KEY_TEXT, config) )
	{
		int_vector<8> text;
		uint64_t num_bytes = 1;
		load_vector_from_file(text, inputfile, num_bytes);
		bool skip = false;
		uint64_t target=0;
		uint64_t sequence_length = 0;
		for(uint64_t i=0; i<text.size(); ++i)
		{
			if(text[i] == '>')
			{
				skip = true;
				if(i>0)
				{
					text[target] = 1;
					++target;
					sequences.emplace_back(sequence_length);
					sequence_length = 0;
				}
			}
			else if(text[i] == '\n')
			{
				skip = false;
			}
			else if(!skip)
			{
				text[target] = text[i];
				++target;
				++sequence_length;
			}
		}
		sequences.emplace_back(sequence_length);
		text.resize(target);
		if(contains_no_zero_symbol(text, inputfile))
		{
			append_zero_symbol(text);
			store_to_cache(text, conf::KEY_TEXT, config);
		}
	}
	else
	{
		config.delete_files = false;
		if(calc_sequences)
		{
			int_vector<8> text;
			load_from_cache(text, conf::KEY_TEXT, config);
			for(uint64_t i=0, len=0; i<text.size(); ++i)
			{
				if(text[i] <= 1)
				{
					sequences.emplace_back(len);
					len = 0;
				}
				else
				{
					++len;
				}
			}
		}
	}
	register_cache_file(conf::KEY_TEXT, config);
	return sequences;
}

bool k_smaller_than_sequences(const vector<uint64_t>& sequences, const string& inputfile, const string& kfilename)
{
	if(!sequences.size())
	{
		cerr << "Could not find a sequence." << endl;
		return false;
	}
	ifstream kfile(kfilename);
	uint64_t k;
	while(kfile >> k)
	{
		for(const auto& length : sequences)
		{
			if(length < k)
			{
				cerr << "k=" << k << " must smaller than sequence length, but in input file '" << inputfile << "' there is a sequence with length " << length << "." << endl;
				return false;
			}
		}
	}
	return true;
}

void create_sa(cache_config& config, const bool fast)
{
	// (2) Check, if the suffix array is cached
	if ( !cache_file_exists(conf::KEY_SA, config) )
	{
		if(!fast)
		{
			construct_config::byte_algo_sa = SE_SAIS;
		}
		construct_sa<8>(config);
	}
	else
	{
		config.delete_files = false;
	}
	register_cache_file(conf::KEY_SA, config);
}

void create_bwt(cache_config& config)
{
	// (3) Check, if bwt is cached
	if ( !cache_file_exists(conf::KEY_BWT, config) )
	{
		construct_bwt<8>(config);
	}
	else
	{
		config.delete_files = false;
	}
	register_cache_file(conf::KEY_BWT, config);
}

void create_da(cache_config& config, const vector<uint64_t>& sequences)
{
	// Load SA
	if ( !cache_file_exists(conf::KEY_SA, config) )
	{
		create_sa(config, true);
	}
	int_vector_buffer<> sa(cache_file_name(conf::KEY_SA, config));

	// Convert length to endpositions
	vector<uint64_t> endpos = sequences;
	for(uint64_t i=1; i<endpos.size(); ++i)
	{
		endpos[i] += (1+endpos[i-1]);
	}

	// Create Document Array
	uint8_t bit_width = 1;
	for(uint64_t j=2; j<sequences.size(); j*=2)
	{
		++bit_width;
	}
	int_vector<> da(sa.size(), 0, bit_width);

	// Fill Document Array
	for(uint64_t i=0; i<sa.size(); ++i)
	{
		uint64_t sa_value = sa[i];
		da[i] = lower_bound(begin(endpos), end(endpos), sa_value)-begin(endpos);
	}

	// Store Document Array
	store_to_cache(da, "DA", config);

	return;
}

uint64_t create_datastructures(cache_config& config, const string& inputfile, const string& kfilename, const bool fast)
{
	// Get input
	vector<uint64_t> sequences = create_text(config, inputfile, true);

	// Check input
	if(!k_smaller_than_sequences(sequences, inputfile, kfilename))
	{
		return 1;
	}

	// Get sa
	create_sa(config, fast);

	// Get bwt
	create_bwt(config);

	return 0;
}
#endif
